/*
 * Disq
 *
 * MIT License
 *
 * Copyright (c) 2018-2019 Disq contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package htsjdk.samtools;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.fail;

import htsjdk.samtools.util.BlockCompressedFilePointerUtil;
import htsjdk.samtools.util.BlockCompressedInputStream;
import htsjdk.tribble.index.IndexFactory;
import htsjdk.tribble.index.tabix.TabixIndex;
import java.io.File;
import java.io.IOException;
import org.junit.Assert;

public class TbiEqualityChecker {

  private File vcfFile;
  private File tbiFile1;
  private File tbiFile2;

  private BlockCompressedInputStream blockStream;

  public static void assertEquals(File vcfFile, File tbiFile1, File tbiFile2, boolean identical)
      throws IOException {
    new TbiEqualityChecker(vcfFile, tbiFile1, tbiFile2).assertEquals(identical);
  }

  public TbiEqualityChecker(File vcfFile, File tbiFile1, File tbiFile2) throws IOException {
    this.vcfFile = vcfFile;
    this.tbiFile1 = tbiFile1;
    this.tbiFile2 = tbiFile2;

    this.blockStream = new BlockCompressedInputStream(vcfFile);
  }

  private void assertEquals(boolean identical) throws IOException {
    TabixIndex tbi1 = (TabixIndex) IndexFactory.loadIndex(tbiFile1.getPath());
    TabixIndex tbi2 = (TabixIndex) IndexFactory.loadIndex(tbiFile2.getPath());

    Assert.assertEquals("Sequences", tbi1.getSequenceNames(), tbi2.getSequenceNames());

    int numReferences = tbi1.getSequenceNames().size();
    for (int i = 0; i < numReferences; i++) {
      BinningIndexContent binningIndexContent1 = tbi1.getIndices()[i];
      BinningIndexContent binningIndexContent2 = tbi2.getIndices()[i];
      assertEquals(binningIndexContent1, binningIndexContent2, identical);
    }
  }

  private void assertEquals(
      BinningIndexContent binningIndexContent1,
      BinningIndexContent binningIndexContent2,
      boolean identical)
      throws IOException {

    assertEquals(binningIndexContent1.getBins(), binningIndexContent2.getBins(), identical);
    assertEquals(binningIndexContent1.getLinearIndex(), binningIndexContent2.getLinearIndex());
  }

  private void assertEquals(
      BinningIndexContent.BinList bins1, BinningIndexContent.BinList bins2, boolean identical)
      throws IOException {
    Assert.assertEquals("Max bin number", bins1.maxBinNumber, bins2.maxBinNumber);
    Assert.assertEquals(
        "Number of non-null bins", bins1.getNumberOfNonNullBins(), bins2.getNumberOfNonNullBins());
    for (int i = 0; i <= bins1.maxBinNumber; i++) {
      assertEquals(bins1.getBin(i), bins2.getBin(i), identical);
    }
  }

  private void assertEquals(Bin bin1, Bin bin2, boolean identical) throws IOException {
    if (bin1 == null || bin2 == null) {
      Assert.assertEquals(bin1, bin2);
      return;
    }
    Assert.assertEquals("Bin number", bin1.getBinNumber(), bin2.getBinNumber());
    Assert.assertEquals("Chunk list size", bin1.getChunkList().size(), bin2.getChunkList().size());
    for (int i = 0; i < bin1.getChunkList().size(); i++) {
      assertEquals(bin1.getChunkList().get(i), bin2.getChunkList().get(i), identical);
    }
  }

  private void assertEquals(LinearIndex linearIndex1, LinearIndex linearIndex2) {
    Assert.assertEquals(
        "Linear index ref",
        linearIndex1.getReferenceSequence(),
        linearIndex2.getReferenceSequence());
    Assert.assertEquals("Linear index size", linearIndex1.size(), linearIndex2.size());
    Assert.assertEquals(
        "Linear index start", linearIndex1.getIndexStart(), linearIndex2.getIndexStart());
    assertArrayEquals(
        "Linear index entries", linearIndex1.getIndexEntries(), linearIndex2.getIndexEntries());
  }

  private void assertEquals(Chunk chunk1, Chunk chunk2, boolean identical) {
    if (identical) {
      Assert.assertEquals("Chunk start", chunk1.getChunkStart(), chunk2.getChunkStart());
      Assert.assertEquals("Chunk end", chunk1.getChunkEnd(), chunk2.getChunkEnd());
    } else {
      assertEquivalent("Chunk start", chunk1.getChunkStart(), chunk2.getChunkStart());
      assertEquivalent("Chunk end", chunk1.getChunkEnd(), chunk2.getChunkEnd());
    }
  }

  private void assertEquivalent(
      String message, long virtualFilePointer1, long virtualFilePointer2) {
    // seek to the given positions check they are actually equivalent
    long norm1 = normalizeVirtualFilePointer(virtualFilePointer1);
    long norm2 = normalizeVirtualFilePointer(virtualFilePointer2);
    Assert.assertEquals(message, norm1, norm2);
  }

  private long normalizeVirtualFilePointer(long virtualFilePointer) {
    try {
      blockStream.seek(virtualFilePointer);
      return blockStream.getFilePointer();
    } catch (IOException e) {
      fail("Failed to seek to " + BlockCompressedFilePointerUtil.asString(virtualFilePointer));
      return -1; // never reached
    }
  }
}
